(* Internal Syntax *)
(* 
 * mainly differs from external syntax in the
 * enforced alternation between type structure
 * and type names
 *)

signature INTSYN =
sig

type tpname = string            (* s, t and nA, nB, nC *)
type label = string             (* l, k *)
type procname = string          (* p, q *)
type varname = string           (* x, y, z *)
type ext = Mark.ext option      (* optional extent = source region info *)

datatype chan = Var of varname              (* x   (statics)  *)
              | Chan of int                 (* (i) (dynamics) *)

datatype tp = Plus of (label * tpname) list (* +{l1:A1,...,ln:An} *)
            | Tensor of tpname * tpname     (* A * B *)
            | One                           (* 1 *)
            | With of (label * tpname) list (* &{l1:A1,...,ln:An *)
            | Lolli of tpname * tpname      (* A -o B *)

datatype value = Channel of chan            (* a *)
               | Label of label * value     (* 'k(V) *)
               | Pair of value * value      (* (V1,V2) *)
               | Unit                       (* () *)
               | MarkedValue of value Mark.marked

datatype proc = Send of chan * value        (* send x V *)
              | Recv of chan * cont        (* recv x K *)
              | Fwd of chan * chan          (* fwd x y *)
              | Call of procname * chan * chan list (* call p x y1...yn *)
              | Cut of chan * tpname option * proc * proc (* x [:A] <- P ; Q *)
              | Marked of proc Mark.marked

     and cont = Cont of (value * proc) list (* (V1 => P1 | ... | VN => Pn) *)

datatype parm = Tp of chan * tpname          (* x : A *)

datatype status = Must          (* must be used *)
                | May           (* may be used or passed on *)
                | Used          (* was used *)
                | Right         (* succedent *)

datatype typing = TpN of chan * tpname * status                    (* x : A{s} *)

type ctx = typing list

datatype def = TpDef of tpname * tp * ext                          (* type t = A *)
             | ProcDef of procname * parm * parm list * proc * ext (* proc p (x:A) (y1:B1)...(yn:Bn) = P *)
             | Exec of procname * ext                              (* exec p *)
             | Fail of def * ext                                   (* fail d *)
             | Error of ext                                        (* error *)

type env = def list

val lookup_procname : env -> procname -> def option
val lookup_tpn : chan -> ctx -> typing option
val expand_opt : env -> tpname -> tp option
val expand : env -> tpname -> tp (* expand env nA requires nA to be defined *)
val select : label -> (label * 'a) list -> 'a option

val pp_tag : label -> string
val pp_chan : chan -> string
val pp_value : value -> string
val is_internal : tpname -> bool
val pp_tpn : env -> tpname -> string   (* expand internal names *)
val pp_tp : env -> tp -> string        (* expand internal names *)
val pp_parm : env -> parm -> string    (* expand internal names *)

type subst = (chan * chan) list  (* a/x or b/a list *)
val subst_proc : subst -> proc -> proc

val abbrev_tp : env -> tp -> tpname option

val subtp : env -> tpname -> tpname -> bool
val eqtp : env -> tpname -> tpname -> bool

end (* signature INTSYN *)


structure IntSyn :> INTSYN =
struct

type tpname = string            (* s, t and nA, nB, nC *)
type label = string             (* l, k *)
type procname = string          (* p, q *)
type varname = string           (* x, y, z *)
type ext = Mark.ext option      (* optional extent = source region info *)

datatype chan = Var of varname              (* x   (statics)  *)
              | Chan of int                 (* (i) (dynamics) *)

datatype tp = Plus of (label * tpname) list (* +{l1:A1,...,ln:An} *)
            | Tensor of tpname * tpname     (* A * B *)
            | One                           (* 1 *)
            | With of (label * tpname) list (* &{l1:A1,...,ln:An *)
            | Lolli of tpname * tpname      (* A -o B *)


datatype value = Channel of chan            (* a *)
               | Label of label * value     (* 'k(V) *)
               | Pair of value * value      (* (V1,V2) *)
               | Unit                       (* () *)
               | MarkedValue of value Mark.marked

datatype proc = Send of chan * value        (* send x V *)
              | Recv of chan * cont        (* recv x K *)
              | Fwd of chan * chan          (* fwd x y *)
              | Call of procname * chan * chan list (* call p x y1...yn *)
              | Cut of chan * tpname option * proc * proc (* x [:A] <- P ; Q *)
              | Marked of proc Mark.marked

     and cont = Cont of (value * proc) list (* (V1 => P1 | ... | VN => Pn) *)

datatype parm = Tp of chan * tpname          (* x : A *)

datatype status = Must          (* must be used *)
                | May           (* may be used or passed on *)
                | Used          (* was used *)
                | Right         (* succedent *)

datatype typing = TpN of chan * tpname * status                    (* x : A{s} *)

type ctx = typing list

datatype def = TpDef of tpname * tp * ext                          (* type t = A *)
             | ProcDef of procname * parm * parm list * proc * ext (* proc p (x:A) (y1:B1)...(yn:Bn) = P *)
             | Exec of procname * ext                              (* exec p *)
             | Fail of def * ext                                   (* fail d *)
             | Error of ext                                        (* error *)

type env = def list

(**********)
(* lookup *)
(**********)

(* lookup_procname env p = SOME(ProcDef(p, ...)) if p is defined in env, NONE otherwise *)
fun lookup_procname ((d as ProcDef(q, _, _, _, _))::env) p =
    if p = q then SOME(d) else lookup_procname env p
  | lookup_procname (_::env) p = lookup_procname env p
  | lookup_procname nil p = NONE

fun lookup_tpn x (TpN(y, nB, s)::Delta) =
    if x = y then SOME(TpN(y, nB, s)) else lookup_tpn x Delta
  | lookup_tpn x nil = NONE

fun expand_opt (TpDef(nB, B, ext)::env) nA =
    if nB = nA then SOME(B) else expand_opt env nA
  | expand_opt (_::env) nA = expand_opt env nA
  | expand_opt nil nA = NONE

fun expand env nA =
    (case expand_opt env nA
      of SOME(A) => A
       | NONE => ( print (nA ^ "\n")
                 ; raise Match ) )

fun select (k:label) ((l,elem)::alts) =
    if k = l then SOME(elem) else select k alts
  | select k nil = NONE

(************)
(* printing *)
(************)

fun pp_tag k = "'" ^ k

fun pp_chan (Var(x)) = x
  | pp_chan (Chan(a)) = "(" ^ Int.toString a ^ ")"

fun pp_value (Channel(a)) = pp_chan a
  | pp_value (Label(k,V)) = pp_tag k ^ " " ^ pp_value V
  | pp_value (Pair(V1,V2)) = "(" ^ pp_value V1 ^ "," ^ pp_value V2 ^ ")"
  | pp_value (Unit) = "()"
  | pp_value (MarkedValue(marked_V)) = pp_value (Mark.data marked_V)

fun is_internal nA = (String.sub(nA, 0) = #"%")

fun parens true str = "(" ^ str ^ ")"
  | parens false str = str

fun pp_tpn env par nA =
    if is_internal nA
    then pp_tp env par (expand env nA)
    else nA
and pp_tp env par (Plus(alts)) = "+" ^ pp_alts env alts
  | pp_tp env par (One) = "1"
  | pp_tp env par (With(alts)) = "&" ^ pp_alts env alts
  | pp_tp env par (Tensor(nA,nB)) = parens par (pp_tpn env true nA ^ " * " ^ pp_tpn env true nB)
  | pp_tp env par (Lolli(nA,nB)) = parens par (pp_tpn env true nA ^ " -o " ^ pp_tpn env true nB)
and pp_alts env alts = "{"
                   ^ String.concatWith ", " (List.map (fn (k,nAk) => pp_tag k ^ " : " ^ pp_tpn env false nAk) alts)
                   ^ "}"

val pp_tpn = fn env => fn nA => pp_tpn env false nA
val pp_tp = fn env => fn A => pp_tp env false A

fun pp_parm env (Tp(x,nA)) = "(" ^ pp_chan x ^ " : " ^ pp_tpn env nA ^ ")"

(****************)
(* substitution *)
(****************)
type subst = (chan * chan) list

(* substitutions are only (x/x) or (a/x) or (a/b)
 * so no capture can occur, only shadowing
 *)

fun id_subst (Channel(x)) theta = (x,x)::theta
  | id_subst (Label(k,V)) theta = id_subst V theta
  | id_subst (Pair(V1,V2)) theta = id_subst V2 (id_subst V1 theta)
  | id_subst (Unit) theta = theta
  | id_subst (MarkedValue(marked_V)) theta = id_subst (Mark.data marked_V) theta                            

fun subst_chan nil x = x
  | subst_chan ((b:chan,y)::theta) x =
    if x = y then b
    else subst_chan theta x

fun subst_value theta (Channel(x)) = Channel(subst_chan theta x)
  | subst_value theta (Label(k,V)) = Label(k, subst_value theta V)
  | subst_value theta (Pair(V1,V2)) = Pair(subst_value theta V1, subst_value theta V2)
  | subst_value theta (Unit) = Unit
  | subst_value theta (MarkedValue(marked_V)) =
    subst_value theta (Mark.data marked_V)

fun subst_proc theta (Send(x,V)) =
    Send(subst_chan theta x, subst_value theta V)
  | subst_proc theta (Recv(x,Cont(K))) =
    Recv(subst_chan theta x, Cont(List.map (subst_cont theta) K))
  | subst_proc theta (Fwd(x,y)) =
    Fwd(subst_chan theta x, subst_chan theta y)
  | subst_proc theta (Call(p, x, ys)) =
    Call(p, subst_chan theta x, List.map (subst_chan theta) ys)
  | subst_proc theta (Cut(x, nAopt, P, Q)) =
    Cut(x, nAopt, subst_proc ((x,x)::theta) P, subst_proc ((x,x)::theta) Q)
  | subst_proc theta (Marked(marked_P)) =
    subst_proc theta (Mark.data marked_P)

and subst_cont theta (V,P) = (V, subst_proc (id_subst V theta) P)

(**********************)
(* abbreviating types *)
(**********************)

fun eq_shallow (Tensor(nA1,nA2)) (Tensor(nB1,nB2)) = nA1 = nB1 andalso nA2 = nB2
  | eq_shallow (One) (One) = true
  | eq_shallow (Lolli(nA1,nA2)) (Lolli(nB1,nB2)) = nA1 = nB1 andalso nA2 = nB2
  | eq_shallow (Plus(alts)) (Plus(blts)) = eq_alts alts blts
  | eq_shallow (With(alts)) (With(blts)) = eq_alts alts blts
  | eq_shallow A B = false
and eq_alts alts blts =
    List.all (fn (k,nAk) => List.exists (fn (l,nBl) => k = l andalso nAk = nBl) blts) alts
    andalso List.all (fn (l,nBl) => List.exists (fn (k,nAk) => l = k andalso nAk = nBl) alts) blts
                                           
fun abbrev_tp env A =
    (case List.find (fn TpDef(nB, B, _) => eq_shallow A B | _ => false) env
      of SOME(TpDef(nB, _, _)) => SOME(nB)
       | NONE => NONE)

(*******************************)
(* subtyping and type equality *)
(*******************************)

fun sub env memo (Tensor(nA1,nA2)) (Tensor(nB1,nB2)) =
    subn env memo nA1 nB1 andalso subn env memo nA2 nB2
  | sub env memo (One) (One) = true
  | sub env memo (Lolli(nA1,nA2)) (Lolli(nB1,nB2)) =
    subn env memo nB1 nA1 andalso subn env memo nA2 nB2
  | sub env memo (Plus(alts)) (Plus(blts)) =
    List.all (fn (k,nAk) => List.exists (fn (l,nBl) => k = l andalso subn env memo nAk nBl) blts) alts
  | sub env memo (With(alts)) (With(blts)) =
    List.all (fn (l,nBl) => List.exists (fn (k,nAk) => l = k andalso subn env memo nAk nBl) alts) blts
  | sub env memo A B = false

and subn env memo nA nB =
    List.exists (fn (nA',nB') => nA = nA' andalso nB = nB') memo
    orelse sub env ((nA,nB)::memo) (expand env nA) (expand env nB)

fun subtp env nA nB = subn env nil nA nB

fun eqtp env nA nB = subn env nil nA nB andalso subn env nil nB nA

end (* structure IntSyn *)
